/***************************************************
 Upright Laser Harp
 Written by Jonathan Bumstead
 21 August 2019
  
  
 /////////////////////////////////////////////////////////
  
  
  
  //////////////////////////////////////////////////////
  Example code for the Adafruit VS1053 Codec Breakout was used. 

  Designed specifically to work with the Adafruit VS1053 Codec Breakout
  ----> https://www.adafruit.com/products/1381

  Written by Limor Fried/Ladyada for Adafruit Industries.
  BSD license, all text above must be included in any redistribution
 ****************************************************/

#include <SoftwareSerial.h>
#include <Stepper.h>

// Motor and Hall setup
#define STEPS 2038 // the number of steps in one revolution
Stepper stepper(STEPS, 8, 10, 9, 11);
bool hE; // current Hall value
bool prevHall; // previous Hall Value
int hEpin = 48; // Hall pin
bool motorCal = 0; // Is motor calibrated
int calVal = -700; // Motor offset from magnet detect
int stepsPerInstrument = 128;


// Midi setup: See http://www.vlsi.fi/fileadmin/datasheets/vs1053.pdf Pg 33
#define VS1053_BANK_MELODY 0x79
#define MIDI_NOTE_ON  0x90
#define MIDI_NOTE_OFF 0x80
#define MIDI_CHAN_MSG 0xB0
#define MIDI_CHAN_BANK 0x00
#define MIDI_CHAN_VOLUME 0x07
#define MIDI_CHAN_PROGRAM 0xC0
int pauseMidi = 10000; //10000
int pauseMidi2 = 50;
SoftwareSerial VS1053_MIDI(0, 2);

// Rotary knob
volatile boolean TurnDetected; // https://forum.arduino.cc/index.php?topic=242356.0
volatile boolean up;

const int PinCLK = 3;                 // Used for generating interrupts using CLK signal
const int PinDT = 4;                  // Used for reading DT signal
const int PinSW = 5;                  // Used for the push button switch

const int maxInstruments = 16;
static long virtualPosition = 1;  // without STATIC it does not count correctly!!!
bool rotSwitch = 1;
bool prevState = 1;
int currentInstrument = 14; // XYLA



// Volume Knob
int aRead;
int volumeValue;
int aPrev = 0;
int aNoise = 1;

byte mapping[120] = {12, 11, 10, 9, 8, 7, 6, 5, 4, 4,  4,4,4,4,4,4,4,4,4,4,  3,3,3,3,3,3,3,3,3,3,   3, 2, 2, 2, 2, 2, 2, 2, 2, 2,   2,2,2,2,2,2,2,2,2,2,   1, 1, 1, 1, 1, 1, 1, 1, 1, 1,    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,    0,0,0,0,0,0,0,0,0,0,  0,0,0,0,0,0,0,0,0,0,   0,0,0,0,0,0,0,0,0,0,   0,0,0,0,0,0,0,0,0,0};

// Lasers and photodiodes
#define numNotes 12
int P[numNotes];
int Lpins[numNotes] = {22, 30, 23, 31, 24, 32, 25, 33, 26, 34, 27, 35};
int Ppins[numNotes] = {A8, A0, A11, A3, A9, A1, A12, A4, A10, A2, A13, A5};
int Pcut = 800;

bool hitValue[numNotes] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
bool prevValue[numNotes] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int notes[numNotes] = {60,72,62,63,64,65,66,67,68,69,70,71}; //{60, 64, 67, 72, 76, 79, 84, 88, 91, 96};
int noteShift = -22;

void setup() {

  pinMode(A14, INPUT);
  Serial.begin(9600);

  VS1053_MIDI.begin(31250); // MIDI uses a 'strange baud rate'

  midiSetChannelBank(0, VS1053_BANK_MELODY);
  midiSetInstrument(0, currentInstrument);
  volUpdate();


  // Initialize rotary button
  pinMode(PinCLK, INPUT);
  pinMode(PinDT, INPUT);
  pinMode(PinSW, INPUT);
  attachInterrupt (digitalPinToInterrupt(PinCLK), isr, FALLING);

  //Initialize motor and Hall
  stepper.setSpeed(15); // 15 rpm
  pinMode(hEpin, INPUT);
  prevHall = digitalRead(hEpin);

  // Lasers
  for (int p = 0; p < numNotes; p++) {
    pinMode(Lpins[p], OUTPUT);
    pinMode(Ppins[p], INPUT);
  }

  
  motorCalF();
  lasersOn(200);
}


void loop() {

  
  volUpdate();
  rotUpdate();


  for (int p = 0; p < numNotes; p++) {
    P[p] = analogRead(Ppins[p]);

    if (P[p] < Pcut) {
      hitValue[p] = 0;
    }
    else if (P[p] >= Pcut) {
      hitValue[p] = 1;
    }


    if (hitValue[p] == 0 && prevValue[p] == 1) //PLAY NOTE //
    {

      midiNoteOn(0, notes[p]+noteShift, 127);
      delayMicroseconds(pauseMidi);
      midiNoteOff(0, notes[p]+noteShift, 127);
      delay(pauseMidi2);

    }

    prevValue[p] = hitValue[p];
    
  }
  
}



//// FUNCTIONS //////

// Volume //
void volUpdate() {
  aRead = analogRead(A14);

  if (aRead > aPrev + aNoise || aRead < aPrev - aNoise) {

    volumeValue = map(aRead, 0, 1024, 0, 120 - 1); //10*mapping[map(aRead, 0, 1024, 0, 29)]; // map(aRead, 0, 1020, 0, 127);
    // Serial.println(10*mapping[volumeValue]);
    midiSetChannelVolume(0, 10*mapping[volumeValue]);//mapping[volumeValue]
    
  }

  aPrev = aRead;

}


// Motor calibration //
void motorCalF() {

  while (motorCal == 0) {
    hE = digitalRead(hEpin);
    //Serial.println(prevHall);

    if (prevHall == 1 && hE == 0) {
      // motor zeroPosition
      motorCal = 1;
      stepper.step(calVal);

    }
    else if (motorCal == 0) { //
      stepper.step(1);
    }

    prevHall = hE;
    delay(1);
  }

}


// Lasers
void lasersOn(int delayTime) {
  for (int p = 0; p < numNotes; p++) {
    digitalWrite(Lpins[p], HIGH);
    delay(delayTime);
  }
}

void lasersOff(int delayTime) {
  for (int p = 0; p < numNotes; p++) {
    digitalWrite(Lpins[p], LOW);
    delay(delayTime);
  }
}









// Rotary //
void rotUpdate() {

  // READ ROTARY AND PICK SOUND
  rotSwitch = digitalRead(PinSW);

  if (rotSwitch == 0 && prevState == 1) {  // check if pushbutton is pressed

    prevState = 0;

  }
  else if (rotSwitch == 1) {
    prevState = 1;
  }

  if (TurnDetected)  {        // do this only if rotation was detected

    //stepper.step(stepsPerInstrument);
    if (up && virtualPosition < maxInstruments) {
      virtualPosition++;
      stepper.step(stepsPerInstrument);
    }
    else if (up && virtualPosition == maxInstruments)
    {
      virtualPosition = 1;
      stepper.step(stepsPerInstrument);
    }
    else if (up == false && virtualPosition > 1) {
      virtualPosition--;
      stepper.step(-stepsPerInstrument);
    }
    else if (up == false && virtualPosition == 1) {
      stepper.step(-stepsPerInstrument);
      virtualPosition = maxInstruments;
    }

    getInstrument(virtualPosition);
    midiSetInstrument(0, currentInstrument);

    TurnDetected = false;          // do NOT repeat IF loop until new rotation detected

  }


}



void isr ()  {                    // Interrupt service routine is executed when a HIGH to LOW transition is detected on CLK
  if (digitalRead(PinCLK))
    up = digitalRead(PinDT);
  else
    up = !digitalRead(PinDT);
  TurnDetected = true;
}


int getInstrument(int virtualPosition) {

  // See http://www.vlsi.fi/fileadmin/datasheets/vs1053.pdf Pg 33
  switch (virtualPosition) {
    case 1:
      currentInstrument = 14; // XYLAPHONE -14
      
      break;
    case 2:
      currentInstrument = 20; // ORGAN - 20
      break;

    case 3:
      currentInstrument = 80; // OCARINA
      break;

    case 4:
      currentInstrument = 103; // ECHOES
      break;

    case 5:
      currentInstrument = 43; // CELLO
      break;

    case 6:
      currentInstrument = 62; // BRASS
      break;

    case 7:
      currentInstrument = 9; // CELESTRA
      break;

    case 8:
      currentInstrument = 10; // GLOCK
      break;

    case 9:
      currentInstrument = 89; // SYNTH
      break;

    case 10:
      currentInstrument = 47; // HARP
      break;

    case 11:
      currentInstrument = 86; // VOICE - 86
      break;

    case 12:
      currentInstrument = 22; // ACCORD
      break;

    case 13:
      currentInstrument = 69; // OBOE
      break;

    case 14:
      currentInstrument = 40; // BASS
      break;

    case 15:
      currentInstrument = 92; // CUSTOM1 - CHOIR
      noteShift = -10;
      break;

    case 16:
      currentInstrument = 104; // CUSTOM2 - SCI FI
      noteShift = -10;
      break;


    default:
      currentInstrument = 14; // XYLA
      break;
  }

  return currentInstrument;

}



//////// MIDI FUNCTIONS //////////
void midiSetInstrument(uint8_t chan, uint8_t inst) {
  if (chan > 15) return;
  inst --; // page 32 has instruments starting with 1 not 0 :(
  if (inst > 127) return;

  VS1053_MIDI.write(MIDI_CHAN_PROGRAM | chan);
  VS1053_MIDI.write(inst);
}


void midiSetChannelVolume(uint8_t chan, uint8_t vol) {
  if (chan > 15) return;
  if (vol > 127) return;

  VS1053_MIDI.write(MIDI_CHAN_MSG | chan);
  VS1053_MIDI.write(MIDI_CHAN_VOLUME);
  VS1053_MIDI.write(vol);
}

void midiSetChannelBank(uint8_t chan, uint8_t bank) {
  if (chan > 15) return;
  if (bank > 127) return;

  VS1053_MIDI.write(MIDI_CHAN_MSG | chan);
  VS1053_MIDI.write((uint8_t)MIDI_CHAN_BANK);
  VS1053_MIDI.write(bank);
}

void midiNoteOn(uint8_t chan, uint8_t n, uint8_t vel) {
  if (chan > 15) return;
  if (n > 127) return;
  if (vel > 127) return;

  VS1053_MIDI.write(MIDI_NOTE_ON | chan);
  VS1053_MIDI.write(n);
  VS1053_MIDI.write(vel);
}

void midiNoteOff(uint8_t chan, uint8_t n, uint8_t vel) {
  if (chan > 15) return;
  if (n > 127) return;
  if (vel > 127) return;

  VS1053_MIDI.write(MIDI_NOTE_OFF | chan);
  VS1053_MIDI.write(n);
  VS1053_MIDI.write(vel);
}
